import logging, nre, os, osproc, sequtils, streams, strtabs, strutils, tables

import private/util
import configuration
from posix import link

type
  Workspace = ref object      ## Data needed by internal build process
    artifactsDir*: string     ## absolute path to the directory for this version
    artifactsRepo*: string    ## absolute path to the global artifacts repo
    buildRef*: string         ## git-style commit reference to the revision we are building
    dir*: string              ## absolute path to the working directory
    env*: StringTableRef      ## environment variables for all build processes
    openedFiles*: seq[File]   ## all files that we have opened that need to be closed
    outputHandler*: HandleProcMsgCB ## handler for process output
    project*: ProjectConfig   ## the project configuration
    projectDef*: ProjectDef   ## the StrawBoss project definition
    status*: BuildStatus      ## the current status of the build
    statusFile*: string       ## absolute path to the build status file
    step*: Step               ## the step we're building
    version*: string          ## project version as returned by versionCmd

proc resolveEnvVars(line: string, env: StringTableRef): string =
  result = line
  for found in line.findAll(re"\$\w+|\$\{[^}]+\}"):
    let key = if found[1] == '{': found[2..^2] else: found[1..^1]
    if env.hasKey(key): result = result.replace(found, env[key])

proc emitStatus(status: BuildStatus, statusFilePath: string,
                outputHandler: HandleProcMsgCB): BuildStatus =
  if statusFilePath != nil: writeFile(statusFilePath, $status)
  if outputHandler != nil:
    outputHandler.sendMsg(status.state & ": " & status.details)
  result = status

proc publishStatus(wksp: Workspace, state, details: string) =
  let status = BuildStatus(state: state, details: details)
  wksp.status = emitStatus(status, wksp.statusFile, wksp.outputHandler)

proc setupProject(wksp: Workspace) =

  # Clone the project into the $temp/repo directory
  let cloneResult = exec("git", wksp.dir,
    ["clone", wksp.projectDef.repo, "repo"],
    wksp.env, {poUsePath}, wksp.outputHandler)

  if cloneResult.exitCode != 0:
    raiseEx "unable to clone repo for '" & wksp.projectDef.name & "'"

  # Checkout the requested ref
  let checkoutResult = exec("git", wksp.dir & "/repo",
    ["checkout", wksp.buildRef],
    wksp.env, {poUsePath}, wksp.outputHandler)

  if checkoutResult.exitCode != 0:
    raiseEx "unable to checkout ref " & wksp.buildRef &
            " for '" & wksp.projectDef.name & "'"

  # Find the strawboss project configuration
  let projCfgFile = wksp.dir & "/repo/" & wksp.projectDef.cfgFilePath
  if not existsFile(projCfgFile):
    raiseEx "Cannot find strawboss project configuration in the project " &
      "repo (expected at '" & wksp.projectDef.cfgFilePath & "')."

  wksp.project = loadProjectConfig(projCfgFile)

  # Merge in the project-defined env vars
  for k, v in wksp.projectDef.envVars: wksp.env[k] = v

  # Get the build version
  let versionProc = startProcess(
    wksp.project.versionCmd,    # command
    wksp.dir & "/repo",         # working dir
    [],                         # args
    wksp.env,                   # environment
    {poUsePath, poEvalCommand}) # options

  let versionResult = waitForWithOutput(versionProc, wksp.outputHandler,
                                        wksp.project.versionCmd)

  if versionResult.exitCode != 0:
    raiseEx "Version command (" & wksp.project.versionCmd &
            ") returned non-zero exit code."

  wksp.version = versionResult.output.strip
  wksp.env["VERSION"] = wksp.version

proc runStep*(wksp: Workspace, step: Step) =

  let SB_EXPECTED_VARS = ["VERSION"]

  wksp.publishStatus("running",
    "running '" & step.name & "' for version " & wksp.version &
    " from " & wksp.buildRef)

  # Ensure all expected environment variables are present.
  for k in (step.expectedEnv & @SB_EXPECTED_VARS):
    if not wksp.env.hasKey(k):
      raiseEx "step " & step.name & " failed: missing required env variable: " & k

  # Ensure that artifacts in steps we depend on are present
  # TODO: detect circular-references in dependency trees.
  for dep in step.depends:
    if not wksp.project.steps.hasKey(dep):
      raiseEx step.name & " depends on " & dep &
              " but there is no step named " & dep
    let depStep = wksp.project.steps[dep]

    # Run that step (may get skipped)
    runStep(wksp, depStep)

    # Add the artifacts directory for the dependent step to our env so that
    # further steps can reference it via $<stepname>_DIR
    wksp.env[depStep.name & "_DIR"] = wksp.artifactsRepo & "/" &
      wksp.project.name & "/" & dep & "/" & wksp.version

  # Run the step command, piping in cmdInput
  wksp.outputHandler.sendMsg step.name & ": starting stepCmd: " & step.stepCmd
  let cmdProc = startProcess(step.stepCmd,
    wksp.dir & "/repo/" & step.workingDir,
    [], wksp.env, {poUsePath, poEvalCommand})

  let cmdInStream = inputStream(cmdProc)

  # Replace env variables in step cmdInput as we pipe it in
  for line in step.cmdInput: cmdInStream.writeLine(line.resolveEnvVars(wksp.env))
  cmdInStream.flush()
  cmdInStream.close()
  
  let cmdResult = waitForWithOutput(cmdProc, wksp.outputHandler, step.stepCmd)

  if cmdResult.exitCode != 0:
    raiseEx "step " & step.name & " failed: step command returned non-zero exit code"

  # Gather the output artifacts (if we have any)
  wksp.outputHandler.sendMsg "artifacts: " & $step.artifacts
  if step.artifacts.len > 0:
    for a in step.artifacts:
      let artifactPath = a.resolveEnvVars(wksp.env)
      let artifactName = artifactPath[(artifactPath.rfind("/")+1)..^1]
      try:
        wksp.outputHandler.sendMsg "copy " & wksp.dir & "/repo/" &
          step.workingDir & "/" & artifactPath & " -> " &
          wksp.artifactsDir & "/" & artifactName

        copyFile(wksp.dir & "/repo/" & step.workingDir & "/" & artifactPath,
                 wksp.artifactsDir & "/" & artifactName)
      except:
        raiseEx "step " & step.name & " failed: unable to copy artifact " &
          artifactPath & ":\n" & getCurrentExceptionMsg()

  wksp.publishStatus("complete", "")

proc runStep*(cfg: StrawBossConfig, req: RunRequest,
              outputHandler: HandleProcMsgCB = nil): BuildStatus =

  result = BuildStatus(
    state: "setup",
    details: "initializing build workspace")
  discard emitStatus(result, nil, outputHandler)

  var wksp: Workspace

  try:
    assert req.workspaceDir.isAbsolute
    if not existsDir(req.workspaceDir): createDir(req.workspaceDir)

    # Find the project definition
    let projectDef = cfg.findProject(req.projectName)

    # Read in the existing system environment
    var env = loadEnv()
    env["GIT_DIR"] = ".git"

    # Setup our STDOUT and STDERR files
    let stdoutFile = open(req.workspaceDir & "/stdout.log", fmWrite)
    let stderrFile = open(req.workspaceDir & "/stderr.log", fmWrite)

    let logFilesOH = makeProcMsgHandler(stdoutFile, stderrFile)

    wksp = Workspace(
      artifactsDir: nil,
      artifactsRepo: cfg.artifactsRepo,
      buildRef:
        if req.buildRef != nil and req.buildRef.len > 0: req.buildRef
        else: projectDef.defaultBranch,
      dir: req.workspaceDir,
      env: env,
      openedFiles: @[stdoutFile, stderrFile],
      outputHandler: combineProcMsgHandlers(outputHandler, logFilesOH),
      project: ProjectConfig(),
      projectDef: projectDef,
      status: result,
      statusFile: req.workspaceDir & "/" & "status.json",
      step: Step(),
      version: nil)

  except:
    when not defined(release): echo getCurrentException().getStackTrace()
    result = BuildStatus(state: "failed",
                         details: getCurrentExceptionMsg())
    try: discard emitStatus(result, nil, outputHandler)
    except: discard ""
    return

  try:
    # Clone the repo and setup the working environment
    wksp.publishStatus("setup",
      "cloning project repo and preparing to run '" & req.stepName & "'")
    wksp.setupProject()

    # Make sure our project directory exists in the artifacts repo
    if not existsDir(wksp.artifactsRepo & "/" & wksp.project.name):
      createDir(wksp.artifactsRepo & "/" & wksp.project.name)

    # Update our cache of project configurations by copying the configuration
    # file to our artifacts directory.
    copyFile(
      wksp.dir & "/repo/" & wksp.projectDef.cfgFilePath,
      cfg.artifactsRepo & "/" & wksp.project.name & "/configuration." &
        wksp.version & ".json")

    # Find the requested step
    if not wksp.project.steps.hasKey(req.stepName):
      raiseEx "no step name '" & req.stepName & "' for " & req.projectName

    var step = wksp.project.steps[req.stepName]

    # Enfore forceRebuild
    if req.forceRebuild: step.dontSkip = true

    # Compose the path to the artifacts directory for this step and version
    wksp.artifactsDir = wksp.artifactsRepo & "/" & wksp.project.name & "/" &
                        step.name & "/" & wksp.version

    # Have we tried to build this before and are we caching the results?
    if existsFile(wksp.artifactsDir & "/status.json") and not step.dontSkip:
      let prevStatus = loadBuildStatus(wksp.artifactsDir & "/status.json")

      # If we succeeded last time, no need to rebuild
      if prevStatus.state == "complete":
        wksp.outputHandler.sendMsg(
          "Skipping step '" & step.name & "' for version '" &
          wksp.version & "': already completed.")
        return prevStatus
      else:
        wksp.outputHandler.sendMsg(
          "Rebuilding failed step '" & step.name & "' for version '" &
          wksp.version & "'.")

    # Make the artifacts directory if it doesn't already exist
    if not existsDir(wksp.artifactsDir): createDir(wksp.artifactsDir)

    # Link status file and output logs to the artifacts dir
    for fn in @["status.json", "stdout.log", "stderr.log"]:
      # TODO: roll old files instead of delete them?
      if existsFile(wksp.artifactsDir & "/" & fn):
        removeFile(wksp.artifactsDir & "/" & fn)

      if link(wksp.dir & "/" & fn, wksp.artifactsDir & "/" & fn) != 0:
        wksp.outputHandler.sendMsg(nil,
          "WARN: could not link " & fn & " to artifacts dir.")

    runStep(wksp, step)

    result = wksp.status

  except:
    when not defined(release): echo getCurrentException().getStackTrace()
    let msg = getCurrentExceptionMsg()
    try:
      wksp.publishStatus("failed", msg)
      result = wksp.status
    except:
      result = BuildStatus(state: "failed", details: msg)
      try: discard emitStatus(result, nil, outputHandler)
      except: discard ""

  finally:
    if wksp != nil:
      for f in wksp.openedFiles:
        try: close(f)
        except: discard ""

